#include "cnrt.h"
#include <iostream>
#include <vector>
#include <fstream>
#include <string>
#include <sstream>
#include "nms.h"
#include <chrono>
#include "cnpy.h"  //prerequire 
#include "macro.h"

void mluNmsThresh(int* result_Index, half* resultScoreClsIdx, int* result_count, half* source);

int main(void){
//------------------------------------------------------------------------------    
//read file  loop : file1 file2 ...
for(int sName = 6; sName < 7; sName++){
////////////////////////////////////////////////////////////////////////////////
//data prepare, use cnpy lib (github cnpy) 
////////////////////////////////////////////////////////////////////////////////
    string scoreFile  = "./data/score/" + std::to_string(sName) + "_score.npy";    
    string boxFile  = "./data/bbox/" + std::to_string(sName) + "_bbox.npy";    
    
    cnpy::NpyArray scores = cnpy::npy_load(scoreFile);
    cnpy::NpyArray bboxs  = cnpy::npy_load(boxFile);
    float* score_data  = scores.data<float>();
    float* bbox_data   = bboxs.data<float>();
    //int* bbox_data   = &((int)bbox_data0);
////////////////////////////////////////////////////////////////////////////////
//chrono timing
////////////////////////////////////////////////////////////////////////////////
    std::chrono::steady_clock::time_point begin, end;
    begin = std::chrono::steady_clock::now();
////////////////////////////////////////////////////////////////////////////////
    cnrtDev_t dev;
    cnrtRet_t ret;
    cnrtInit(0);
    cnrtGetDeviceHandle(&dev,0);
    cnrtSetCurrentDevice(dev);
    cnrtQueue_t pQueue;
    cnrtCreateQueue(&pQueue);
    cnrtDim3_t dim;
    dim.x=16;
    dim.y=1;
    dim.z=1;
//    cnrtFunctionType_t c = CNRT_FUNC_TYPE_BLOCK;
    cnrtFunctionType_t c = CNRT_FUNC_TYPE_UNION4; 

////////////////////////////////////////////////////////////////////////////////
//convert input score data from float to half
////////////////////////////////////////////////////////////////////////////////
    half * input_score_half = (half*)(malloc(N_box * Cls * sizeof(half)));
    for( int i =0; i< N_box*Cls; i++){
	    cnrtConvertFloatToHalf(&input_score_half[i] ,score_data[i]);

    }
////////////////////////////////////////////////////////////////////////////////
//MLU memory allocate , cp input from host to MLU
////////////////////////////////////////////////////////////////////////////////
    uint32_t*  mlu_outIndex=NULL;
    half*      mlu_out=NULL;
    uint32_t*  mlu_out_count=NULL;
    half*      mlu_in=NULL;

    cnrtMalloc((void**)&mlu_outIndex,N_box * sizeof(int));
    cnrtMalloc((void**)&mlu_out, N_box * 2 *  sizeof(half));
    cnrtMalloc((void**)&mlu_out_count, dim.x * sizeof(int));
    cnrtMalloc((void**)&mlu_in, N_box * Cls * sizeof(half));

    cnrtMemcpy(mlu_in, input_score_half, N_box  * Cls * sizeof(half), CNRT_MEM_TRANS_DIR_HOST2DEV);

////////////////////////////////////////////////////////////////////////////////
//MLU timing , MLU init , MLU invoke
////////////////////////////////////////////////////////////////////////////////
    cnrtNotifier_t notifier_begin;
    cnrtNotifier_t notifier_end;
    cnrtCreateNotifier(&notifier_begin);
    cnrtCreateNotifier(&notifier_end);
    float timeTotal = 0.0;
    cnrtPlaceNotifier(notifier_begin,pQueue);


    cnrtKernelInitParam_t init_param;
    CNRT_CHECK(cnrtCreateKernelInitParam(&init_param));
    CNRT_CHECK(cnrtInitKernelMemory((const void*)mluNmsThresh, init_param)); 

    cnrtKernelParamsBuffer_t params;
    cnrtGetKernelParamsBuffer(&params);

    cnrtKernelParamsBufferAddParam(params,&mlu_outIndex,sizeof(int* ));
    cnrtKernelParamsBufferAddParam(params,&mlu_out,sizeof(half *));
    cnrtKernelParamsBufferAddParam(params,&mlu_out_count,sizeof(int*));
    cnrtKernelParamsBufferAddParam(params,&mlu_in, sizeof(half *));
    
    unsigned int affinity1 = 0x00;
    cnrtInvokeParam_t invoke_param1;
    invoke_param1.invoke_param_type = CNRT_INVOKE_PARAM_TYPE_0;
    invoke_param1.cluster_affinity.affinity = &affinity1;
    
    CNRT_CHECK(cnrtInvokeKernel_V3((void *)&mluNmsThresh,init_param,dim,params,c,pQueue,(void*)&invoke_param1)); 
    
    cnrtPlaceNotifier(notifier_end,pQueue);
    ret = cnrtSyncQueue(pQueue);
    cnrtNotifierDuration(notifier_begin, notifier_end, &timeTotal);
    printf("Hardware Tottal time : %.3f ms\n",timeTotal/1000.0);
//////////////////////////////////////////////////////////////////////////////// 
//kernal func result from MLU to Host, convert dataType, Free memory
////////////////////////////////////////////////////////////////////////////////
    uint32_t h_index[N_box];
    half     h_scoreCls_half[N_box*2];
    uint32_t h_counts[dim.x];

    ret = cnrtMemcpy((void**)&h_index, mlu_outIndex, N_box *  sizeof(uint32_t),CNRT_MEM_TRANS_DIR_DEV2HOST);
    ret = cnrtMemcpy((void**)&h_scoreCls_half, mlu_out, N_box * 2* sizeof(half),CNRT_MEM_TRANS_DIR_DEV2HOST);
    ret = cnrtMemcpy((void**)&h_counts, mlu_out_count, dim.x *  sizeof(uint32_t), CNRT_MEM_TRANS_DIR_DEV2HOST); 

    float h_score[N_box];
    int h_Cls[N_box];
    for(int i= 0;i< N_box; i++){
	    cnrtConvertHalfToFloat(&h_score[i], h_scoreCls_half[i*2]);
	    h_Cls[i] = *((uint16_t * ) (h_scoreCls_half + i*2 + 1));
    }
    ret = cnrtDestroyQueue(pQueue);
    cnrtFree(mlu_out_count);
    cnrtFree(mlu_in);
    cnrtFree(mlu_out);
    cnrtFree(mlu_outIndex);
    //cnrtFreeHost(input_score_half);
    cnrtDestroy();
////////////////////////////////////////////////////////////////////////////////
//
    int count_all=0;
    for(int i =0; i< dim.x; i++)
	    count_all  += h_counts[i];
    
    printf("image  %d : countALL = %d\n",sName, count_all);


    Solution a;
    Bbox inputItem;
    vector<Bbox> input;
    for(int j = 0;j<dim.x;j++){
        for(int i=0; i < h_counts[j]; i ++){
//	printf("j = %d and hcounts [j] = %d\n", j, h_counts[j]);
  	inputItem.score = h_score[j*ROUND + i];	
	inputItem.Idx = h_Cls[j*ROUND + i];
//	printf("i  = %d and hidx  = %d and score is %f \n", i, inputItem.Idx,inputItem.score);
	inputItem.x1 = bbox_data[4*h_index[j*ROUND + i]];	
	inputItem.y1 = bbox_data[4*h_index[j*ROUND + i]+1];	
	inputItem.x2 = bbox_data[4*h_index[j*ROUND + i]+2];	
	inputItem.y2 = bbox_data[4*h_index[j*ROUND + i]+3];	
	input.push_back(inputItem);
        }
    }
    std::vector<Bbox> res;
    res = a.nms(input,0.45);
    
    printf("---------sName  = %d ----------\n",sName);
    for(int i = 0; i< res.size();i++){
	    printf("%f %f %f %f %d  %f",res[i].x1, res[i].y1, res[i].x2, res[i].y2, res[i].Idx, res[i].score);
	    cout<< endl;
    }
    
//    free(result2H);
 //   free(result2HIdx);
 
    end = std::chrono::steady_clock::now();
    std::cout <<" Total time used is : "<< (float)std::chrono::duration_cast<std::chrono::microseconds>(end - begin).count()/1e3<<"[ms]"<<std::endl;
    
    }
    return 0;
}
